/* $Id: bpfBucket.h,v 1.12.2.2 2003/03/19 02:57:18 lergnom Exp $ -*- C++ -*-
 * 
 * libbpfl - Bayesian Probability Filtering Library
 * Copyright (C) 2003 Anthony Ventimiglia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */
#ifndef BPF_BUCKET_H
#define BPF_BUCKET_H

#ifdef __BPF_BUILD
#include "common.h"
#include "bpfHash_map.h"
#else
#include <libbpfl/bpfHash_map.h>
#endif

#include <fstream>
#include <string>

/// The entire BPFL Library will be in namespace bpf.
namespace bpf {

///This is the class used to store Token data for the Filters.
///The class or type TokenType used to instantiate the template has a
///few requirements:
///  - It has to have a equality operator ( bool operator== )
///  - It has to have an input operator ( std::istream& operator>> )
///  - It has to hace an output operator ( std::ostream& operator<< )
///  - You have to write a hash function that fits the Bucket::hash
///    Prototype. 
template <typename TokenType, typename HashObject>
class Bucket 
{
public:
    Bucket ( const std::string& name );
    virtual ~Bucket () {}

    ///Returns the Token's conditional probability P(D|H).
    ///This is used by the Filter class. The score is the total number
    ///of times the token is in the Bucket divided by the total number
    ///of tokens in the bucket. 
    virtual double score ( const TokenType& );

    ///Add another token to the bucket.
    virtual void add ( const TokenType&, const unsigned long &c = 1 );

    ///Write the Bucket to a file. 
    virtual void save ( const std::string& );

    ///Load the Bucket from a file.
    virtual void load ( const std::string& );

    const std::string name;

private:
    typedef Hash_map<TokenType, unsigned long, HashObject> table_t;
    Hash_map<TokenType, unsigned long, HashObject> table;
    mutable unsigned int token_count;

};


template <typename T, typename H>
Bucket<T, H>::Bucket ( const std::string& BucketName )
    : name ( BucketName )
{
    token_count = 0;
}
    
template<typename T, typename H>
double Bucket<T, H>::score ( const T& tok ) 
{
    typename table_t::iterator i = table.find ( tok );
    return double ( *i ) / double ( token_count );
}

template<typename T, typename H>
void Bucket<T, H>::add ( const T& tok, const unsigned long& c )
{
    typename table_t::iterator i = table.find ( tok );
    i += c;
    token_count += c;
}

template<typename T, typename H>
void Bucket<T, H>::save ( const std::string& fname )
{
    std::ofstream fout ( fname.c_str() );
    for ( typename table_t::iterator i = table.begin(); i != table.end(); 
	  ++i )
	fout << i.key() << ' ' << *i << '\n';
    fout.close();
}

template<typename T, typename H>
void Bucket<T, H>::load ( const std::string& fname )
{
    std::ifstream fin ( fname.c_str() );
    while ( fin ) {
	T tok;
	unsigned long int val;
	fin >> tok >> val;
	if ( !fin ) break;
	typename table_t::iterator ti = table.find ( tok );
	ti =  val;
	token_count += val;
    }
    fin.close();
}



};				// namespace bpf

#endif // BPF_BUCKET_H
