/* $Id: bpfHash_map.h,v 1.13.2.2 2003/03/19 03:23:07 lergnom Exp $ -*- C++ -*-
 * 
 * libbpfl - Bayesian Probability Filtering Library
 * Copyright (C) 2003 Anthony Ventimiglia
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */
#ifndef BPF_HASH_MAP_H
#define BPF_HASH_MAP_H

#ifdef __BPF_BUILD
#include "common.h"
#include "Hash_map_base.h"
#else
#include <libbpfl/Hash_map_base.h>
#endif

#include <cstdlib>

namespace bpf {

///This is used internally by the Bucket class to store tokens.
///For a Key type to be used to fill this template, it must have:
///  - an equality operator ( operator== )
///  - A hash function object, that returns a Hash value with the operator()
///
///The ValueType must have a copy operator, that allows it to be set using
///'=' ( ValueType& operator= ( const ValueType& ))
///
template <typename KeyType, typename ValueType, typename HashObject>
class Hash_map : public Hash_map_base
{
public:

    typedef HashObject hasher;
    
    ///Used to access individual cells, and for sequential iteration
    typedef class iterator;

    Hash_map();
    virtual ~Hash_map();
    
    ///Find the cell in Table with a Key matching k.
    ///Returns a Hash_map::iterator that can be used to set the value.
    ///If a Key matching k was not in the hash it is inserted and it's
    ///iterator is returned.
    const iterator find ( const KeyType& k );
    
    ///Begin sequential iteration.
    const iterator begin ();

    ///Next element in sequntial iteration 
    const iterator next ( const iterator& i );

    ///End of sequential Iteration.
    ///Thes esequential iterator methods should be used like the the STL
    ///iterators, but they aren't yet that nice to use.
    const iterator end () { return iterator( *this, size()); }

private:
    tablesize size_index;
    tablesize cells_used;

    typedef struct {
	bool in_use;
	KeyType key;
	ValueType value;
    } cell;

    cell* table;
    hasher hash;

    tablesize size () { return size_lookup_table[size_index]; }
    void resize ();

    // Clear all cells in table.
    void clear ();
};

// Include iterator class
#ifdef __BPF_BUILD
#include "Hash_map_iterator.h"
#else 
#include <libbpfl/Hash_map_iterator.h>
#endif

// Implementation for template dependant methods follow.

template <typename K, typename V, typename H>
Hash_map<K, V, H>::Hash_map () : Hash_map_base()
{
    size_index = 0;
    cells_used = 0;
    table = new cell[size()];
    clear();
}

template <typename K, typename V, typename H>
Hash_map<K, V, H>::~Hash_map ()
{
    delete[] table;
}

template <typename K, typename V, typename H>
void Hash_map<K, V, H>::clear ()
{
    for ( tablesize i = 0; i < size(); ++i ) 
	table[i].in_use = false;
}

template <typename K, typename V, typename H>
const typename Hash_map<K, V, H>::iterator 
Hash_map<K, V, H>::find ( const K& k )
{
    tablesize hash_val = hash ( k ) % size();
    tablesize ind = hash_val;

    ///@todo Add bounds checking to make sure the hash() function is working
    ///      properly
    while ( table[ind].in_use && !(table[ind].key == k) ) {
	ind = (1 + ind) % size();
	if ( ind == hash_val ) {
	    ///@todo throw an exception here
	    std::exit ( 1 );
	}
    }
    if ( !table[ind].in_use ) {
	if ( (cells_used) > (2 * size() / 3)) {
	    resize();
	    return iterator ( find ( k )) ;
	} else {
	    table[ind].in_use = true;
	    table[ind].key = k;
	    cells_used++;
	}
    }
    return iterator ( *this, ind );
}

template <typename K, typename V, typename H>
void Hash_map<K, V, H>::resize ()
{
    Hash_map::cell* old_table = table;
    tablesize old_size = size();
    if ( ++(size_index) == Hash_map::lookup_table_size ) {
	///@todo throw exception
	std::exit (1);
    }
    table = new Hash_map::cell[size()];
    clear();
    for ( tablesize oi = 0; oi < old_size; ++oi ) 
	if ( old_table[oi].in_use ) {
	    iterator i = find ( old_table[oi].key );
	    i = old_table[oi].value;
	}
    delete[] old_table;
}

template <typename K, typename V, typename H>
const typename Hash_map<K, V, H>::iterator Hash_map<K, V, H>::begin ()
{
    tablesize i;
    for ( i = 0; i < size() && !(table[i].in_use); ++i );
    return iterator ( *this, i );
}

 

};				// namespace bpf

#endif // BPF_HASH_MAP_H
